/*
  poi_icon_select.cpp
  Copyright (C) 2008-2010 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include "qgmap.h"
#include <qstring.h>
#include <qlabel.h>
#include <qvbox.h>
#include <qhbox.h>
#include <qpushbutton.h>
#include <qsizepolicy.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qpixmap.h>
#include <qmessagebox.h>
#include <qfileinfo.h>
#include <qcombobox.h>
#include <qspinbox.h>
#include <qlistbox.h>

#include <qiconview.h>

#include "poi_db.h"
#include "poi_icon_select.h"
#include "poi_icon_dialog.h"

#define DIALOG_WIDTH  380

PoiIconSelect::PoiIconSelect( QWidget* parent, const char* name,
								bool modal, WFlags fl)
	: QDialog( parent, name, modal, fl )
{
	if (g_machine.resolution() == QVGA) {
		resize(DIALOG_WIDTH / 2, DIALOG_WIDTH / 2);
	} else {
		resize(DIALOG_WIDTH, DIALOG_WIDTH);
	}
	resize(size().expandedTo(minimumSizeHint()));

	setCaption(QString::fromUtf8("アイコン選択"));

	QFont small_font = font();
	small_font.setPointSize(small_font.pointSize() * 3 / 4);

	QVBoxLayout *vbox = new QVBoxLayout(this, 0, 3);

	icon_view = new QIconView(this);
	vbox->addWidget(icon_view);

	icon_view->setResizeMode(QIconView::Adjust);
	icon_view->setSelectionMode(QIconView::Single);
	icon_view->setItemsMovable(false);
	icon_view->setFont(small_font);

	QHBox *hbox = new QHBox(this);
	vbox->addWidget(hbox);

	QPushButton *add_button = new QPushButton(QString::fromUtf8("追加"), hbox);
	QPushButton *del_button = new QPushButton(QString::fromUtf8("削除"), hbox);
#ifndef QWS
	QPushButton *cancel = new QPushButton("Cancel", hbox);
	QPushButton *ok = new QPushButton("OK", hbox);
#endif

	connect(icon_view, SIGNAL(doubleClicked(QIconViewItem *)),
			this, SLOT(decideIcon(QIconViewItem *)));
	connect(icon_view, SIGNAL(returnPressed(QIconViewItem *)),
			this, SLOT(decideIcon(QIconViewItem *)));

	connect(add_button, SIGNAL(clicked()), this, SLOT(press_add_button()));
	connect(del_button, SIGNAL(clicked()), this, SLOT(press_del_button()));
#ifndef QWS
	connect(cancel, SIGNAL(clicked()), this, SLOT(reject()));
	connect(ok, SIGNAL(clicked()), this, SLOT(accept()));
#endif

	updateIcon();

	show();
	updateGeometry();
	QSize size = (g_machine.screensize() - this->size()) / 2;
	move(size.width(), size.height());
}

PoiIconSelect::~PoiIconSelect()
{
}

int PoiIconSelect::getIconId()
{
	QIconViewItem *item = icon_view->currentItem();
	if (!item)
		return -1;

	return icon_id_table[icon_view->index(item)];
}

void PoiIconSelect::updateIcon(int id)
{
	PoiDb db(g_db_poi);

	PoiDbIcon *icon = db.queryPoiIcon();
	if (!icon)
		return;

	icon_view->clear();
	if (!icon->next()) {
		delete icon;
		return;
	}

	icon_id_table.resize(16);

	int i = 0;
	do {
		QPixmap pixmap = QPixmap(icon->getIconData());
		if (!pixmap.isNull()) {
			if (i >= (int)icon_id_table.size())
				icon_id_table.resize(i + 16);

			icon_id_table[i] = icon->getId();
			QString name = icon_id_table[i] < PoiDb::PRESET_OFFSET ?
				QString::number(icon_id_table[i]):
				QString("P%1").arg(icon_id_table[i] - PoiDb::PRESET_OFFSET);

			QIconViewItem *item = new QIconViewItem(icon_view, name, pixmap);
			if (id == icon->getId())
				icon_view->setCurrentItem(item);
			i++;
		}
	} while(icon->next());

	delete icon;
}

void PoiIconSelect::decideIcon(QIconViewItem *item)
{
	UNUSED(item);
	accept();
}

void PoiIconSelect::press_add_button()
{
	PoiIconDialog dlg(this);

	if (dlg.exec() == QDialog::Accepted) {
		int id = dlg.getIconId();
		updateIcon(id);
	}
}

void PoiIconSelect::press_del_button()
{
	int id = getIconId();
	if (id < 0)
		return;
	if (id >= PoiDb::PRESET_OFFSET) {
		QMessageBox::critical(this, 
							  "Error!!",
							  QString::fromUtf8("プリセットアイコンは削除できません"));
		return;
	}

	if (QMessageBox::warning(this, QString::fromUtf8("削除の確認"),
							 QString::fromUtf8("アイコンを削除してよろしいですか?"),
							 QMessageBox::Cancel | QMessageBox::Default,
							 QMessageBox::Ok | QMessageBox::Escape) != QMessageBox::Ok) {
		return;
	}

	PoiDb db(g_db_poi);
	db.delIcon(id);
	updateIcon();
}
