/*
  pos.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <math.h>
#include <qt.h>
#include <qregexp.h>
#include <qstringlist.h>
#include <qstring.h>
#include "pos.h"
#include "qgmap.h"

Pos::Pos(double _lat, double _lon, int _zoom)
{
	setZoom(_zoom);
	setLatLon(_lat, _lon);
}

Pos::Pos(int _x, int _y, int _zoom)
{
	setZoom(_zoom);
	setPos(_x, _y);
}

void Pos::setPos(int _x, int _y)
{
	x = _x;
	y = _y;
}

void Pos::setZoom(int _zoom)
{
	zoom = _zoom;
}

void Pos::setLatLon(double lat, double lon, bool wgs)
{
	int max_len = IMAGE_SIZE << (19-zoom);
	double dx, dy;

	while (lon < -180.0 || lon >= 180.0)
		lon = lon + ((lon<0.0)?360.0:-360.0);

	while (lat < -90.0 || lat >= 90.0)
		lat = lat + ((lat<0.0)?180.0:-180.0);

	if(!wgs){
		double wlat,wlon;
		wlat = lat - 0.000106950 * lat + 0.000017464 * lon + 0.0046017;
		wlon = lon - 0.000046038 * lat - 0.000083043 * lon + 0.010040;
		lat = wlat;
		lon = wlon;
	}

	dx = (lon+180.0)/360.0*(double)max_len;
	dy = (0.5-logf(tanf((lat+90.0)/180.0*3.141592/2.0))/(2.0*3.141592))
		*(double)max_len;

	x = (int)dx;
	y = (int)dy;
}

Pos &Pos::offset(int _x, int _y)
{
	static Pos retpos;

	retpos.setZoom(zoom);
	retpos.setPos(this->getX() + _x, this->getY() + _y);

	return retpos;
}

Pos &Pos::offset_clip(int _x, int _y)
{
	static Pos retpos;
	int max;

	max = (1<<(19-zoom))*IMAGE_SIZE;
	_x += this->getX();
	_y += this->getY();
	if(_x < 0) _x = 0;
	if(_y < 0) _y = 0;
	if(_x >= max) _x = max-1;
	if(_y >= max) _y = max-1;

	retpos.setZoom(zoom);
	retpos.setPos(_x, _y);

	return retpos;
}

double Pos::getLat() const
{
	int zm = 1 << (19-zoom);
	double lat;

	lat = 90.0 - atan(pow(0.367879, (0.5-(double)y/((double)zm*256.0))*2.0*3.141592))
		/3.141592*180.0*2.0;

	return lat;
}

double Pos::getLon() const
{
	int max_len = IMAGE_SIZE << (19-zoom);
	double lon;
	int _x = x;

	while(_x < 0)
		_x += max_len;
	_x %= max_len;

	lon = (double)_x/(double)max_len*360.0-180.0;

	return lon;
}

double Pos::getRadLat() const
{
	int zm = 1 << (19-zoom);
	double lat;

	lat = PI / 2 - atan(pow(0.367879, (0.5-(double)y/((double)zm * 256.0)) * 2.0 * PI))
		* 2.0;

	return lat;
}

double Pos::getRadLon() const
{
	int max_len = IMAGE_SIZE << (19-zoom);
	double lon;
	int _x = x;

	while(_x < 0)
		_x += max_len;
	_x %= max_len;

	lon = (double)_x / (double)max_len * 2.0 * PI - PI;

	return lon;
}


int Pos::getX() const
{
	return x;
}

int Pos::getY() const
{
	return y;
}

int Pos::getZoom() const
{
	return zoom;
}

bool Pos::zoomup(int _zoom)
{
	int new_zoom;
	int d;

	new_zoom = zoom + _zoom;
	if(new_zoom > 19) new_zoom = 19;
	if(new_zoom < 0) new_zoom = 0;

	if(new_zoom == zoom)
		return false;

	if(_zoom < 0){
		// zoom in
		d = 1 << (-_zoom);
		x *= d;
		y *= d;
	}else{
		d = 1 << _zoom;
		x /= d;
		y /= d;
	}
	zoom = new_zoom;

	return true;
}

bool Pos::zoomTo(int _zoom)
{
	if(_zoom != zoom)
		return zoomup(_zoom - zoom);
	return false;
}

void Pos::normalize()
{
	int max;

	max = (1<<(19-zoom))*IMAGE_SIZE;

	while(x<0)
		x+=max;
	x%=max;
}

void Pos::clipY()
{
	int max;

	max = (1<<(19-zoom))*IMAGE_SIZE;

	if(y < 0) y = 0;
	if(y > max) y = max;
}

#define EARTH_B2PA2 ((EARTH_RB / EARTH_RA) * (EARTH_RB / EARTH_RA))

#define EARTH_2 ((1.0 - EARTH_B2PA2) / 2.0)
#define EARTH_1 (1.0 - EARTH_2)

double Pos::getLatLength()
{
	double lat, latlen;
	double rad;

	lat = getLat();

	rad = lat * PI / 180.0;

	latlen = 2.0*PI*((double)EARTH_RA)*cos(rad)
		/sqrt(EARTH_1+EARTH_2*cos(2.0*rad));

	return latlen;
}

double Pos::getDotLength()
{
	unsigned long int mapsize = IMAGE_SIZE << (19-zoom);
	double latlen=getLatLength();

	latlen /= (double)mapsize;

	return latlen;
}

int Pos::getScaleSize(int max, int *mul)
{
	double dotlen = getDotLength();
	int maxlen = (int)(dotlen * max);

	*mul=1;
	while(maxlen >= 10){
		*mul *= 10;
		maxlen /= 10;
	}

	if(maxlen>=5)
		*mul *= 5;
	else if(maxlen >= 2)
		*mul *= 2;

	return (int)((double)*mul/dotlen);
}

bool Pos::setPos(QString const &str, bool wgs)
{
	QStringList list;
	QString strlat, strlon;
	double lat,lon;

	list = QStringList::split(QRegExp("\\s*,\\s*"), str);
	if(list.count()<2 || list.count()>3)
		return false;

	strlat = list[0].stripWhiteSpace();
	strlon = list[1].stripWhiteSpace();

	// parse Lat, Lon
	if(!parseStr(strlat, &lat))
		return false;
	if(!parseStr(strlon, &lon))
		return false;

	// check Lat, Lon
	if(lat > 85.0 || lat < -85.0)
		return false;
	if(lon > 180.0 || lon < -180.0)
		return false;

	// if zoom is specified
	if(list.count() == 3){
		QString zoomstr = list[2].stripWhiteSpace();
		// zoom is integer?
		if(zoomstr.contains(QRegExp("[^\\d]")) != 0)
			return false;

		// check zoom
		int zoom = zoomstr.toInt();
		if(zoom < 0 || zoom > 19)
			return false;

		setZoom(zoom);
	}

	setLatLon(lat, lon, wgs);

	return true;
}

bool Pos::parseStr(const QString str, double *value)
{
	QRegExp integer("^\\d\\d*$");
	QRegExp dms("^[+-]?\\d?\\d?\\d\\.[0-5]?\\d\\.[0-5]?\\d");
	QRegExp deg("^[+-]?\\d?\\d?\\d\\.?\\d*$");
	bool ok;

	if(integer.match(str)==0){
		*value = str.toDouble();
	}else if(deg.match(str)==0){

		*value = str.toDouble(&ok);
		if(!ok) return false;

	}else if(dms.match(str)==0){
		QString d,m,s,work;
		int i;
		i = str.find('.');
		d = str.left(i);
		work = str.mid(i+1);

		i = work.find('.');
		m = work.left(i);
		s = work.mid(i+1);

		*value = d.toDouble(&ok);
		if(!ok) return false;
		*value += m.toDouble(&ok)/60.0;
		if(!ok) return false;
		*value += s.toDouble(&ok)/3600.0;
		if(!ok) return false;

	}else{

		return false;

	}

	return true;
}

QString Pos::serialize() const
{
	QString str;

	str.sprintf("%d,%d,%d", x, y, zoom);

	return str;
}

bool Pos::deserialize(QString const &str)
{
	QString sx,sy,szoom,work;
	int i;

	i = str.find(',');
	if(i < 0)
		return false;
	sx = str.left(i);
	work = str.mid(i+1);

	i = work.find(',');
	if(i < 0)
		return false;
	sy = work.left(i);
	szoom = work.mid(i+1);

	x = sx.toInt();
	y = sy.toInt();
	zoom = szoom.toInt();

	return true;
}

bool Pos::operator==(const Pos &pos) const
{
	return x == pos.x && y == pos.y && zoom == pos.zoom;
}

bool Pos::operator!=(const Pos &pos) const
{
	return !(*this == pos);
}

QString Pos::getPos()
{
	QString str;

	str.sprintf("%f,%f,%d", getLat(), getLon(), zoom);

	return str;
}

QPoint Pos::getPoint()
{
	return QPoint(x, y);
}

void Pos::convTokyoToWGS(double *wgs_lat, double *wgs_lon,
								   double tokyo_lat, double tokyo_lon)
{
	double lat, lon;
	lat = tokyo_lat - 0.000106950 * tokyo_lat
		+ 0.000017464 * tokyo_lon + 0.0046017;
	lon = tokyo_lon - 0.000046038 * tokyo_lat
		- 0.000083043 * tokyo_lon + 0.010040;

	*wgs_lat = lat;
	*wgs_lon = lon;
}

double Pos::distance(const Pos &pos)
{
#define E2 ((EARTH_RA * EARTH_RA - EARTH_RB * EARTH_RB) / \
			(EARTH_RA * EARTH_RA))	/* 第一離心率^2 */

	double lat, pos_lat;
	double lon, pos_lon;
	double P, dP, dR, M, N, W, D, x, y;
	double sinP, cosP;

	lat = getRadLat();
	lon = getRadLon();
	pos_lat = pos.getRadLat();
	pos_lon = pos.getRadLon();

	P = (lat + pos_lat) / 2.0;
	dP = lat - pos_lat;
	dR = lon - pos_lon;
	if (dR < PI)
		dR += PI * 2.0;
	if (dR > PI)
		dR -= PI * 2.0;

	sinP = sin(P);
	cosP = cos(P);
	W = sqrt(1.0 - E2 * sinP * sinP);
	N = EARTH_RA / W;
	M = N * (1.0 - E2) / W / W;

	x = M * dP;
	y = N * cosP * dR;
	D = sqrt(x * x + y * y);

	return D;
}

double Pos::direction_deg(const Pos &pos)
{
	double deg;
	Pos p1, p2;;
	int x;
	const int map_max = IMAGE_SIZE << 19;
	p1 = *this;
	p2 = pos;
	p1.zoomTo(0);
	p2.zoomTo(0);

	x = p2.x - p1.x;
	if (x <= -map_max / 2)
		x += map_max;
	if (x > map_max / 2)
		x -= map_max;

	deg = atan2(x, -(p2.y - p1.y)) / PI * 180.0;
	if (deg < 0.0)
		deg += 360.0;

	return deg;
}

QString Pos::direction_str(const Pos &pos)
{
	const char *dir[] = {
		"北", "北北東", "北東", "東北東", "東", "東南東", "南東", "南南東",
		"南", "南南西", "南西", "西南西", "西", "西北西", "北西", "北北西",
	};
	int deg;

	deg = (direction_deg(pos) + 360.0/32.0) / (360.0 / 16.0);
	deg = (deg + 16) % 16;
	printf("deg: %lf -> %d\n", direction_deg(pos), deg);
	printf("%s\n", dir[deg]);

	return QString::fromUtf8(dir[deg]);
}
